#togglestate {
    display: none;
}
div#custom-version-navigation .navigation {
    font-family: monospace;
    padding: 0 0 0 1.5rem;
}
div#custom-version-navigation .dropdown {
    display: inline-block;
    width: calc(100% - 1.5rem);
    height: 100%;
    padding-left: 0%;
}
div#custom-version-navigation button.dropdown {
    border: none;
    text-align: left;
    background-color: inherit;
    width: 1.5rem;
    text-align: center;
}
div#custom-version-navigation button.dropdown:hover {
    background-color: var(--nav-hover);
}
div#custom-version-navigation .subversion {
    display: none;
    padding-left: 2rem;
}
div.notice-marker {
    position: fixed;
    bottom: 0;
    right: 0;
    width: calc(100% - var(--sidebar-width));
    max-width: 100%;
    padding: .4rem 5rem .4rem 1rem;
    border: solid 3px red;
    background-color: var(--active);
    font-weight: bold;
}
div#custom-version-navigation summary {
    position: inherit;
    font-size: inherit;
    list-style-position: inherit;
    padding-left: 1.7rem;
    margin-left: 1.3rem;
}
div#custom-version-navigation summary:hover {
    background-color: var(--nav-hover);
}
h1.modulename {
    display: inline;
}