{% extends "default/module.html.jinja2" %}

{% if env["PDOC_DEFINE_VIEW_SOURCE_MACRO"] == "1" %}
    {% defaultmacro view_source(doc) %}
        {{ view_source_state(doc) }}
        {{ view_source_button(doc) }}
        {{ view_source_code(doc) }}
        <br />
    {% enddefaultmacro %}
{% endif %}

{% block title %}
    {% if env["BUILD_DOCS_LEVEL"] == "homepage" %}
screen_brightness_control Documentation
    {% elif env["BUILD_DOCS_LEVEL"] == "extras" %}
{{ module.modulename.split(".")[-1] }} - screen_brightness_control
    {% else %}
{{ super() }}
    {% endif %}
{% endblock %}

{% block head %}
    {{ super() }}
    <script type="text/javascript">
        {% include "version_navigator_loader.js.jinja2" %}
    </script>
{% endblock %}

{% block module_list_link %}
    {% if env["BUILD_DOCS_LEVEL"] != "homepage" %}
        <a class="pdoc-button module-list-button" href="https://crozzers.github.io/screen_brightness_control">
            {% include "resources/box-arrow-in-left.svg" %}
            &nbsp;
            Back to homepage
        </a>
    {% endif %}
    {% if env["BUILD_DOCS_LEVEL"] != "extras" %}
        {{ super() }}
    {% endif %}
{% endblock %}

{% block nav_footer %}

    <div id="custom-version-navigation">
    </div>

    <h2>
        <a href="https://github.com/Crozzers/screen_brightness_control">GitHub</a>
    </h2>
    <ul>
        <li><a href="https://github.com/Crozzers/screen_brightness_control/releases">Releases</a></li>
        <li><a href="https://github.com/Crozzers/screen_brightness_control/issues">Issue Tracker</a></li>
        <li><a href="https://github.com/Crozzers/screen_brightness_control/pulls">Pull Requests</a></li>
    </ul>
    <h2>
        <a href="https://pypi.org/project/screen-brightness-control/">PyPI</a>
    </h2>
    <h2>
        <a href="https://github.com/Crozzers/screen_brightness_control/blob/main/LICENSE">License</a>
    </h2>
    {{ super() }}
{% endblock %}

{% block module_info %}
    <section>
        {% block edit_button %}
            {{ super() }}
        {% endblock %}

        {% if env["BUILD_DOCS_LEVEL"] == "api" %}
            {{ module_name() }}
        {% elif env["BUILD_DOCS_LEVEL"] == "extras" %}
            <h1 class="modulename">
            {{ module.modulename.split(".")[-1] }}
            </h1>
        {% endif %}
        {{ docstring(module) }}
        {{ view_source(module) }}
    </section>
{% endblock %}